#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <sys/time.h>

#include <cblas.h>

#include <assert.h>

#include "msg.h"
#include "hw_sched.h"
#include "mrses_impl.h"

enum MRSESEntriesT {
    MRSES_RUN_ENTRY = 0,
    MRSES_ITERATE_ENTRY = 1
};
typedef enum MRSESEntriesT MRSESEntries;

MRSESContext mrses_create_context() {
    MRSESContext ctx;
    posix_memalign((void*)&ctx, HW_ALIGN, calc_alloc(sizeof(MRSESContextS), HW_ALIGN));
    if (ctx) {
        memset(ctx, 0, sizeof(MRSESContextS));
    }
    return ctx;
}

int mrses_init_context(MRSESContext ctx, int properties, int nA, const MRSESDataType *A, int nB, const MRSESDataType *B) {
    int err = 0;
    int i;
    int nAB;

    int alloc;
    int palloc;

#ifdef HW_HAVE_SPU
    int aA, aB;
#endif /* HW_HAVE_SPU */

    int matrix_size;
    MRSESDataType *ones, *meanA, *meanB;

    assert(ctx);
    assert(A);
    assert(B);
    assert(nA > 0);
    assert(nB > 0);
    assert(properties > 0);

    mrses_free_context(ctx);

    nAB = max(nA, nB);    
    alloc = calc_alloc(nAB * sizeof(MRSESDataType), HW_ALIGN) / sizeof(MRSESDataType);
    palloc = calc_alloc(properties * sizeof(MRSESDataType), HW_ALIGN) / sizeof(MRSESDataType);
    matrix_size = properties * alloc;
    
    ctx->nA = nA;
    ctx->nB = nB;
    ctx->alloc = alloc;
    ctx->properties = properties;    
    ctx->palloc = calc_alloc(properties * sizeof(MRSESIntType), HW_ALIGN) / sizeof(MRSESIntType);

    posix_memalign((void*)&ctx->A, HW_ALIGN, (alloc * properties)*sizeof(MRSESDataType));
    posix_memalign((void*)&ctx->B, HW_ALIGN, (alloc * properties)*sizeof(MRSESDataType));
    posix_memalign((void*)&ctx->mean, HW_ALIGN, palloc * sizeof(MRSESDataType));
    posix_memalign((void*)&meanB, HW_ALIGN, properties*sizeof(MRSESDataType));
    posix_memalign((void*)&ones, HW_ALIGN, nAB*sizeof(MRSESDataType));

    meanA = ctx->mean;

    if ((!ctx->A)||(!ctx->B)||(!meanA)||(!meanB)||(!ones)) {
	if (ones) free(ones);
	if (meanB) free(meanB);
	return 1;
    }
	
    for (i = 0; i < nAB; i++) 
	ones[i] = 1.;
    
    memset(ctx->A, 0x7F, alloc * properties * sizeof(MRSESDataType));
//    blas_gemv(CblasRowMajor,CblasNoTrans, properties, nA, 1. / nA, A, nA, ones, 1, 0, meanA, 1);
//    blas_gemv(CblasRowMajor,CblasNoTrans, properties, nB, 1. / nB, B, nB, ones, 1, 0, meanB, 1);

    for (i = 0; i < properties; i++) {
	memcpy(ctx->A + i * alloc, A + i * nA, nA * sizeof(MRSESDataType));
    }

#ifdef HW_HAVE_SPU
    aA = calc_alloc(nA * sizeof(MRSESDataType), HW_ALIGN) - nA * sizeof(MRSESDataType);
    if (aA) {
	for (i = 0; i < properties; i++) {
	    memset(ctx->A + i * alloc + nA, 0, aA);
	}
    }
#endif /* HW_HAVE_SPU */

    blas_gemv(CblasRowMajor,CblasNoTrans, properties, nA, 1. / nA, ctx->A, alloc, ones, 1, 0, meanA, 1);
    for (i = 0; i < nA; i++) {
	blas_axpy(properties, -1, meanA, 1, ctx->A + i, alloc);
    }
    blas_scal(matrix_size, 1.0 / sqrt(nA), ctx->A, 1);


    for (i = 0; i < properties; i++) {
	memcpy(ctx->B + i * alloc, B + i * nB, nB * sizeof(MRSESDataType));
    }

#ifdef HW_HAVE_SPU
    aB = calc_alloc(nB * sizeof(MRSESDataType), HW_ALIGN) - nB * sizeof(MRSESDataType);
    if (aB) {
	for (i = 0; i < properties; i++) {
	    memset(ctx->B + i * alloc + nB, 0, aB);
	}
    }
#endif /* HW_HAVE_SPU */

    blas_gemv(CblasRowMajor,CblasNoTrans, properties, nB, 1. / nB, ctx->B, alloc, ones, 1, 0, meanB, 1);
    for (i = 0; i < nB; i++) {
	blas_axpy(properties, -1.0, meanB, 1, ctx->B + i, alloc);
    }
    blas_scal(matrix_size, 1.0 / sqrt(nB), ctx->B, 1);

    blas_axpy(properties, -1.0, meanB, 1, meanA, 1);


/*
    blas_gemv(CblasRowMajor,CblasNoTrans, properties, nA, 1. / nA, A, nA, ones, 1, 0, meanA, 1);
    blas_gemv(CblasRowMajor,CblasNoTrans, properties, nB, 1. / nB, B, nB, ones, 1, 0, meanB, 1);

    matrix_size = properties * nA;
    memcpy(ctx->A, A, matrix_size * sizeof(MRSESDataType));
    for (i = 0; i < nA; i++) {
	blas_axpy(properties, -1, meanA, 1, ctx->A + i, nA);
    }
    blas_scal(matrix_size, 1.0 / sqrt(nA), ctx->A, 1);

    matrix_size = properties * nB;
    memcpy(ctx->B, B, matrix_size * sizeof(MRSESDataType));
    for (i = 0; i < nB; i++) {
	blas_axpy(properties, -1.0, meanB, 1, ctx->B + i, nB);
    }
    blas_scal(matrix_size, 1.0 / sqrt(nB), ctx->B, 1);
    blas_axpy(properties, -1.0, meanB, 1, meanA, 1);
*/    

/*
    //printf("%f %f.\n", meanA[4], meanA[5]);
    //printf("%f %f.\n", meanB[4], meanB[5]);
*/



    ctx->sched = hw_sched_create();
    hw_sched_set_sequential_mode(ctx->sched, &ctx->block_size, &ctx->cur_chunk);

    free(ones);
    free(meanB);
    
    return err;
}

int mrses_init_context_from_double(MRSESContext ctx, int properties, int nA, const double *A, int nB, const double *B) {
    int err = 0;
    int i, j;
    int nAB;

    int alloc;
    int palloc;

#ifdef HW_HAVE_SPU
    int aA, aB;
#endif /* HW_HAVE_SPU */

    int matrix_size;
    MRSESDataType *ones, *meanA, *meanB;

    assert(ctx);
    assert(A);
    assert(B);
    assert(nA > 0);
    assert(nB > 0);
    assert(properties > 0);

    mrses_free_context(ctx);

    nAB = max(nA, nB);    
    alloc = calc_alloc(nAB * sizeof(MRSESDataType), HW_ALIGN) / sizeof(MRSESDataType);
    palloc = calc_alloc(properties * sizeof(MRSESDataType), HW_ALIGN) / sizeof(MRSESDataType);
    matrix_size = properties * alloc;
    
    ctx->nA = nA;
    ctx->nB = nB;
    ctx->alloc = alloc;
    ctx->properties = properties;    
    ctx->palloc = calc_alloc(properties * sizeof(MRSESIntType), HW_ALIGN) / sizeof(MRSESIntType);

    posix_memalign((void*)&ctx->A, HW_ALIGN, (alloc * properties)*sizeof(MRSESDataType));
    posix_memalign((void*)&ctx->B, HW_ALIGN, (alloc * properties)*sizeof(MRSESDataType));
    posix_memalign((void*)&ctx->mean, HW_ALIGN, palloc * sizeof(MRSESDataType));
    posix_memalign((void*)&meanB, HW_ALIGN, properties*sizeof(MRSESDataType));
    posix_memalign((void*)&ones, HW_ALIGN, nAB*sizeof(MRSESDataType));

    meanA = ctx->mean;

    if ((!ctx->A)||(!ctx->B)||(!meanA)||(!meanB)||(!ones)) {
	if (ones) free(ones);
	if (meanB) free(meanB);
	return 1;
    }
	
    for (i = 0; i < nAB; i++) 
	ones[i] = 1.;
    
    memset(ctx->A, 0x7F, alloc * properties * sizeof(MRSESDataType));


    for (i = 0; i < properties; i++) {
	/* DS: Change1, 
	memcpy(ctx->A + i * alloc, A + i * nA, nA * sizeof(MRSESDataType));
	*/
	for (j = 0; j < nA; j++) {
	    ctx->A[i * alloc + j] = A[i * nA + j];
	}
    }

#ifdef HW_HAVE_SPU
    aA = calc_alloc(nA * sizeof(MRSESDataType), HW_ALIGN) - nA * sizeof(MRSESDataType);
    if (aA) {
	for (i = 0; i < properties; i++) {
	    memset(ctx->A + i * alloc + nA, 0, aA);
	}
    }
#endif /* HW_HAVE_SPU */

    blas_gemv(CblasRowMajor,CblasNoTrans, properties, nA, 1. / nA, ctx->A, alloc, ones, 1, 0, meanA, 1);
    for (i = 0; i < nA; i++) {
	blas_axpy(properties, -1, meanA, 1, ctx->A + i, alloc);
    }
    blas_scal(matrix_size, 1.0 / sqrt(nA), ctx->A, 1);


    for (i = 0; i < properties; i++) {
	/* DS: Change2 
	    memcpy(ctx->B + i * alloc, B + i * nB, nB * sizeof(MRSESDataType));
	*/
	for (j = 0; j < nA; j++) {
	    ctx->B[i * alloc + j] = B[i * nA + j];
	}
    }

#ifdef HW_HAVE_SPU
    aB = calc_alloc(nB * sizeof(MRSESDataType), HW_ALIGN) - nB * sizeof(MRSESDataType);
    if (aB) {
	for (i = 0; i < properties; i++) {
	    memset(ctx->B + i * alloc + nB, 0, aB);
	}
    }
#endif /* HW_HAVE_SPU */

    blas_gemv(CblasRowMajor,CblasNoTrans, properties, nB, 1. / nB, ctx->B, alloc, ones, 1, 0, meanB, 1);
    for (i = 0; i < nB; i++) {
	blas_axpy(properties, -1.0, meanB, 1, ctx->B + i, alloc);
    }
    blas_scal(matrix_size, 1.0 / sqrt(nB), ctx->B, 1);

    blas_axpy(properties, -1.0, meanB, 1, meanA, 1);


/*
    blas_gemv(CblasRowMajor,CblasNoTrans, properties, nA, 1. / nA, A, nA, ones, 1, 0, meanA, 1);
    blas_gemv(CblasRowMajor,CblasNoTrans, properties, nB, 1. / nB, B, nB, ones, 1, 0, meanB, 1);

    matrix_size = properties * nA;
    memcpy(ctx->A, A, matrix_size * sizeof(MRSESDataType));
    for (i = 0; i < nA; i++) {
	blas_axpy(properties, -1, meanA, 1, ctx->A + i, nA);
    }
    blas_scal(matrix_size, 1.0 / sqrt(nA), ctx->A, 1);

    matrix_size = properties * nB;
    memcpy(ctx->B, B, matrix_size * sizeof(MRSESDataType));
    for (i = 0; i < nB; i++) {
	blas_axpy(properties, -1.0, meanB, 1, ctx->B + i, nB);
    }
    blas_scal(matrix_size, 1.0 / sqrt(nB), ctx->B, 1);
    blas_axpy(properties, -1.0, meanB, 1, meanA, 1);
*/    

/*
    //printf("%f %f.\n", meanA[4], meanA[5]);
    //printf("%f %f.\n", meanB[4], meanB[5]);
*/



    ctx->sched = hw_sched_create();
    hw_sched_set_sequential_mode(ctx->sched, &ctx->block_size, &ctx->cur_chunk);

    free(ones);
    free(meanB);
    
    return err;
}



int mrses_set_distance_mode(MRSESContext ctx, MRSESDistance dist) {
    assert(ctx); 
    assert(dist < MRSES_DISTANCE_LAST);

    ctx->dist = dist;
    return 0;
}

int mrses_prepare(MRSESContext ctx, int width, int block_size) {
#ifdef HW_HAVE_SPU
    int at_once, pack;
    int index_block;
#endif /* HW_HAVE_SPU */

    assert(ctx);

    ctx->width = width;

#ifdef HW_HAVE_SPU
# ifdef HW_USE_BLOCKED_MULTIPLY
    at_once = SPE_BLOCK / width;
    if (at_once < 1) at_once = 1;
# else /* HW_USE_BLOCKED_MULTIPLY */
    at_once = 1;
#endif /* HW_USE_BLOCKED_MULTIPLY */

    pack = SIMD_BLOCK / sizeof(MRSESDataType);
    at_once *= pack;
        
    index_block = HW_ALIGN / min(sizeof(MRSESDataType), sizeof(MRSESIntType));

    ctx->iterate_size = lcm3(at_once, index_block, HW_ITERATE_BLOCKS);
#else /* HW_HAVE_SPU */
    ctx->iterate_size = HW_ITERATE_BLOCKS;
#endif /* HW_HAVE_SPU */

    reportMessage("block size: %i", ctx->iterate_size);

    ctx->max_block_size = calc_alloc(block_size, ctx->iterate_size);
    
    return 0;
}

void mrses_free_context(MRSESContext ctx) {
    assert(ctx);

    if (ctx->sched) hw_sched_destroy(ctx->sched);
    
    if (ctx->index) free(ctx->index);
    if (ctx->mean) free(ctx->mean);
    if (ctx->B) free(ctx->B);
    if (ctx->A) free(ctx->A);
    
    memset(ctx, 0, sizeof(MRSESContextS));
}

void mrses_destroy_context(MRSESContext ctx) {
    mrses_free_context(ctx);
    free(ctx);
}


int mrses_compute(MRSESContext ctx, int block_size, MRSESIntType *index, MRSESDataType *res) {
    assert(ctx);
    assert(block_size > 0);
    assert(block_size <= ctx->max_block_size);
    assert(index);
    assert(res);
    assert(!ctx->index);

    ctx->block_size = block_size;
    ctx->cur_chunk = 0;
    ctx->index = index;
    ctx->result = res;
    
    hw_sched_schedule_task(ctx->sched, ctx, MRSES_RUN_ENTRY);
    hw_sched_wait_task(ctx->sched);

    ctx->index = NULL;
    
    return 0;
}

int mrses_iterate(MRSESContext ctx, int iterations, int block_size, MRSESIntType *ires) {
    int i, width, properties, palloc;
#ifndef FIX_RANDOM
    struct timeval tv;
#endif /* FIX_RANDOM */

    MRSESIntType *index, *iptr, *iend; 
    
    assert(ctx);
    assert(iterations > 0);
    assert(block_size > 0);
    assert(block_size <= ctx->max_block_size);

    block_size = calc_alloc(block_size, ctx->iterate_size);

    width = ctx->width;
    properties = ctx->properties;
    palloc = ctx->palloc;

#ifdef FIX_RANDOM
    srandom(FIX_RANDOM);
#else /* FIX_RANDOM */
    gettimeofday(&tv, NULL);
    srandom(tv.tv_usec);
#endif /* FIX_RANDOM */

    if (ctx->index) {
        if (!ires) hw_sched_wait_task(ctx->sched);
	index = ctx->index;
    } else {
        posix_memalign((void*)&index, HW_ALIGN, (block_size * palloc)*sizeof(MRSESIntType));
	if (!index) {
	    return 1;
	}
    }

    for (i = 0; i < properties; i++) {
	index[i] = i;
    }

    iend = index + block_size * palloc;    
    for (iptr = index + palloc; iptr < iend; iptr += palloc) {
	memcpy(iptr, index, properties * sizeof(MRSESIntType));
	
	for (i = 0; i < width; i++) {
	    SWAPi(iptr, properties, i);
	}
    }
    for (i = 0; i < width; i++) {
	SWAPi(index, properties, i);
    }

/*
    for (i = 0; i < 5; i++) {
	int j
	for (j = 0; j < 5; j++) {
	    printf("%5i ", index[i * properties + j]);
	}
	printf("\n");
    }
    printf("\n");
*/
    
    ctx->block_size = block_size / ctx->iterate_size;
    ctx->cur_chunk = 0;

    ctx->iterations = iterations;
    ctx->index = index;
    ctx->ires = ires;
    ctx->result = NULL;

    hw_sched_schedule_task(ctx->sched, ctx, MRSES_ITERATE_ENTRY);

    if (ires) {
	hw_sched_wait_task(ctx->sched);
	free(ctx->index); ctx->index = NULL;
    }

    return 0;
}

int mrses_get_results(MRSESContext ctx, uint32_t *hist) {
    int i, j;
    
    HWSched sched = ctx->sched;
    HWThread *thr = sched->thread;
    int n_threads = sched->n_threads;
    int properties = ctx->properties;
    
    uint32_t *thr_hist = NULL;

    if (ctx->index) {
        hw_sched_wait_task(ctx->sched);
	free(ctx->index); ctx->index = NULL;
    }

    for (i = 0; i < n_threads; i++) {
	thr_hist = (uint32_t*)(thr[i]->data);
	if (thr_hist) break;
    }
    
    if (thr_hist) {
	memcpy(hist, thr_hist, properties * sizeof(uint32_t));
	memset(thr_hist, 0, properties * sizeof(uint32_t));
    }
    
    for (++i; i < n_threads; i++) {
	thr_hist = (uint32_t*)(thr[i]->data);
	if (!thr_hist) continue;
	
	for (j = 0; j < properties; j++) {
	    hist[j] += thr_hist[j];
	}
	memset(thr_hist, 0, properties * sizeof(uint32_t));
    }
    
    return 0;
}
